#include "../nestypes.h"
#include "../audiosys.h"
#include "../handler.h"
#include "../nes6502.h"
#include "logtable.h"
#include "s_apu.h"

#define NES_BASECYCLES (21477270)

/* 31 - 7 - log2(MAX_FREQ/60) > SPF_BITS */
/* MAX_FREQ:96000 13.3 > SPF_BITS */
#define SPF_BITS 12

/* 31 - log2(NES_BASECYCLES/(12*MIN_FREQ)) > CPS_BITS  */
/* MIN_FREQ:11025 23.6 > CPS_BITS */
/* 32-12(max spd) > CPS_BITS */
#define CPS_BITS 19

#define ENV_BITS	(31 - CPS_BITS - 5/* bitsof env(1-16) */)
#define SWEEP_BITS	(31 - CPS_BITS - 4/* bitsof sweep(1-8) */)

#define KEY_ON		1
#define KEY_RELEASE 2

typedef struct {
	Uint32 cps;
	Int32 cycles;
	Int32 sweepphase;
	Int32 envphase;

	Uint32 spd;
	Uint32 envspd;
	Uint32 sweepspd;

	Uint32 length;
	Uint32 freq;
	Uint32 mastervolume;
	Uint32 release;

	Uint8 regs[4];
	Uint8 update;
	Uint8 key;
	Uint8 adr;
	Uint8 envadr;
	Uint8 duty;
	Uint8 mute;
} APU_SQUARE;
typedef struct {
	Uint32 cps;
	Int32 cycles;

	Uint32 spd;

	Uint32 length;
	Uint32 length2;
	Uint32 freq;
	Uint32 mastervolume;
	Uint32 release;

	Uint8 regs[4];
	Uint8 update;
	Uint8 key;
	Uint8 adr;
	Uint8 counter_started;
	Uint8 write_latency;
	Uint8 mute;
} APU_TRIANGLE;
typedef struct {
	Uint32 cps;
	Int32 cycles;
	Int32 envphase;

	Uint32 spd;
	Uint32 envspd;

	Uint32 length;
	Uint32 freq;
	Uint32 mastervolume;
	Uint32 rng;
	Int32 rngadr;

	Uint8 regs[4];
	Uint8 update;
	Uint8 key;
	Uint8 envadr;
	Uint8 mute;
} APU_NOISE;
typedef struct {
	Uint32 cps;
	Int32 cycles;

	Uint32 spd;

	Uint32 length;
	Uint32 mastervolume;
	Uint32 adr;
	Int32 output;

	Uint8 regs[4];
	Uint8 update;
	Uint8 key;
	Uint8 irq_report;
	Uint8 input;
	Uint8 first;
	Uint8 mute;
} APU_DPCM;
typedef struct {
	APU_SQUARE square[2];
	APU_TRIANGLE triangle;
	APU_NOISE noise;
	APU_DPCM dpcm;
	Uint8 regs[0x20];
} APUSOUND;

typedef APU_SQUARE MMC5_SQUARE;
typedef struct {
	Int32 output;
	Uint8 key;
	Uint8 mute;
} MMC5_DA;

typedef struct {
	MMC5_SQUARE square[2];
	MMC5_DA da;
} MMC5SOUND;

/* ------------------------- */
/*  NES INTERNAL SOUND(APU)  */
/* ------------------------- */

#include "sct_main.h"

static APUSOUND apu;

#define V(x) (x*64/60)
static Uint vbl_length[32] =
{
	V(0x05), V(0x7F), V(0x0A), V(0x01), V(0x14), V(0x02), V(0x28), V(0x03),
	V(0x50), V(0x04), V(0x1E), V(0x05), V(0x07), V(0x06), V(0x0D), V(0x07),
	V(0x06), V(0x08), V(0x0C), V(0x09), V(0x18), V(0x0A), V(0x30), V(0x0B),
	V(0x60), V(0x0C), V(0x24), V(0x0D), V(0x08), V(0x0E), V(0x10), V(0x0F),
};
#undef V
/*
odd 127,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15

even
5,5 * 2,
20,20 * 2,
80,30  160??  32? (80 * 2) & 0x7F
7,7 * 2
6,6 * 2
24,24*2
96,   192?? 36???
8,8*2
*/
#define V(x) ((x) << CPS_BITS)
static const Uint32 spd_limit[8] =
{
	V(0x3FF), V(0x555), V(0x666), V(0x71C), 
	V(0x787), V(0x7C1), V(0x7E0), V(0x7F0),
};
static Uint32 noise_freq[16] =
{
	 V(   4), V(   8), V(  16), V(  32),
	 V(  64), V(  96), V( 128), V( 160),
	 V( 202), V( 254), V( 380), V( 508),
	 V( 762), V(1016), V(2034), V(2046),
};
static Uint32 dpcm_freq[16] =
{
	V(428), V(380), V(340), V(320),
	V(286), V(254), V(226), V(214),
	V(190), V(160), V(142), V(128),
	V(106), V( 85), V( 72), V( 54),
};
#undef V


static Int32 APUSoundSquareRender(APU_SQUARE *ch)
{
	Uint32 output;
	if (ch->update)
	{
		if (ch->update & 1)
		{
			ch->duty = (ch->regs[0] >> 4) & 0x0C;
			if (ch->duty == 0) ch->duty = 2;
			ch->envspd = ((ch->regs[0] & 0x0F) + 1) << (CPS_BITS + ENV_BITS);
		}
		if (ch->update & 2)
		{
			ch->sweepspd = (((ch->regs[1] >> 4) & 0x07) + 1) << (CPS_BITS + SWEEP_BITS);
		}
		if (ch->update & (4 | 8))
		{
			ch->spd = (((ch->regs[3] & 7) << 8) + ch->regs[2] + 1) << CPS_BITS;
		}
		if ((ch->update & 8) && (ch->key & KEY_ON))
		{
			ch->key &= ~KEY_RELEASE;
			ch->length = (vbl_length[ch->regs[3] >> 3] * ch->freq) >> 6;
			ch->envadr = 0;
		}
		ch->update = 0;
	}

	if (ch->key == 0) return 0;

	ch->envphase -= ch->cps >> (13 - ENV_BITS);
	if (ch->regs[0] & 0x20)
	{
		while (ch->envphase < 0)
		{
			ch->envphase += ch->envspd;
			ch->envadr++;
		}
		ch->envadr &= 0x0F;
	}
	else
	{
		while (ch->envphase < 0)
		{
			ch->envphase += ch->envspd;
			ch->envadr += (ch->envadr < 15);
		}
	}

	if (ch->length)
	{
		if (!(ch->regs[0] & 0x20)) ch->length--;
	}
	else
	{
		ch->key |= KEY_RELEASE;
	}

	if ((ch->regs[1] & 0x80))
	{
		if (ch->regs[1] & 7)
		{
			ch->sweepphase -= ch->cps >> (14 - SWEEP_BITS);
			if (ch->regs[1] & 8)
				while (ch->sweepphase < 0)
				{
					ch->sweepphase += ch->sweepspd;
					ch->spd -= ch->spd >> (ch->regs[1] & 7);
				}
			else
				while (ch->sweepphase < 0)
				{
					ch->sweepphase += ch->sweepspd;
					ch->spd += ch->spd >> (ch->regs[1] & 7);
				}
		}
	}

	if (ch->spd < (4 << CPS_BITS)) return 0;
	if (!(ch->regs[1] & 8))
	{
		if (ch->spd > spd_limit[ch->regs[1] & 7]) return 0;
	}

	ch->cycles -= ch->cps;
	while (ch->cycles < 0)
	{
		ch->cycles += ch->spd;
		ch->adr++;
	}
	ch->adr &= 0x0F;

	if (ch->key & KEY_RELEASE)
	{
		if (ch->release < (31 << (LOG_BITS + 1)))
			ch->release += 3 << (LOG_BITS - 8 + 1);
	}
	else
	{
		ch->release = 0;
	}

	if (ch->mute) return 0;

	if (ch->regs[0] & 0x10) /* fixed volume */
		output = ch->regs[0] & 0x0F;
	else
		output = 15 - ch->envadr;

	output = LinearToLog(output) + ch->mastervolume + ch->release;
	output += (ch->adr < ch->duty);
	return LogToLinear(output, LOG_LIN_BITS - LIN_BITS - 16);
}

static Int32 APUSoundTriangleRender(APU_TRIANGLE *ch)
{
	Uint32 output;
	if (ch->update)
	{
		if (ch->update & (4 | 8))
		{
			ch->spd = (((ch->regs[3] & 7) << 8) + ch->regs[2] + 1) << CPS_BITS;
		}
		ch->update = 0;
	}

	if (ch->key == 0) return 0;

	if (!ch->counter_started && !(ch->regs[0] & 0x80))
	{
		if (ch->write_latency) ch->write_latency--;
		if (!ch->write_latency) ch->counter_started = 1;
	}
	if (ch->counter_started)
	{
		if (ch->length2) ch->length2--;
		if (ch->length)
		{
			if (!(ch->regs[0] & 0x80)) ch->length--;
		}
		else
		{
			ch->key |= KEY_RELEASE;
		}
	}
	if (!ch->length2) ch->key |= KEY_RELEASE;

	if (ch->spd < (4 << CPS_BITS)) return 0;
/*
	if (ch->regs[3] & 8)
	{
		if (ch->spd > spd_limit[ch->regs[1] & 7]) return 0;
	}
*/

	ch->cycles -= ch->cps;
	while (ch->cycles < 0)
	{
		ch->cycles += ch->spd;
		ch->adr++;
	}
	ch->adr &= 0x1F;

	if (ch->key & KEY_RELEASE)
	{
		if (ch->release < (31 << (LOG_BITS + 1)))
			ch->release += 3 << (LOG_BITS - 8 + 1);
	}
	else
	{
		ch->release = 0;
	}

	if (ch->mute) return 0;

	output = (ch->adr & 7) << 1;
	if (ch->adr & 0x08) output = 0x10 - output;

	output = LinearToLog(output) + ch->mastervolume + ch->release;
	output += (ch->adr < 0x10);
	return LogToLinear(output, LOG_LIN_BITS - LIN_BITS - 16);
}

static Int32 APUSoundNoiseRender(APU_NOISE *ch)
{
	Int32 output;

	if (ch->update)
	{
		if (ch->update & 1)
		{
			ch->envspd = ((ch->regs[0] & 0x0F) + 1) << (CPS_BITS + ENV_BITS);
		}
		if (ch->update & 4)
		{
			ch->spd = noise_freq[ch->regs[2] & 0x0F];
		}
		if ((ch->update & 8) && ch->key)
		{
			ch->length = (vbl_length[ch->regs[3] >> 3] * ch->freq) >> 6;
			ch->envadr = 0;
		}
		ch->update = 0;
	}
	if (!ch->key) return 0;

	ch->envphase -= ch->cps >> (13 - ENV_BITS);
	if (ch->regs[0] & 0x20)
	{
		while (ch->envphase < 0)
		{
			ch->envphase += ch->envspd;
			ch->envadr++;
		}
		ch->envadr &= 0x0F;
	}
	else
	{
		while (ch->envphase < 0)
		{
			ch->envphase += ch->envspd;
			ch->envadr += (ch->envadr < 15);
		}
	}

	if (!ch->length) return 0;
	if (!(ch->regs[0] & 0x20)) ch->length--;

	/* if (ch->spd < (4 << CPS_BITS)) return 0; */

	ch->cycles -= ch->cps;
	if (ch->regs[2] & 0x80)
	{
		while (ch->cycles < 0)
		{
			Uint32 rngx;
			ch->cycles += ch->spd;
			rngx  = ch->rng & 1;
			ch->rng >>= 1;
			rngx ^= ch->rng & 1;
			ch->rng |= rngx << 3;
		}
	}
	else
	{
		while (ch->cycles < 0)
		{
			Uint32 rngx;
			ch->cycles += ch->spd;
			if (--ch->rngadr < 0)
			{
				ch->rng = 0x11;
				ch->rngadr = 0x3FFF;
			}
			rngx  = ch->rng & 1;
			ch->rng >>= 1;
			rngx ^= ch->rng & 1;
			ch->rng |= rngx << 12;
		}
	}

	if (ch->mute) return 0;

	if (ch->regs[0] & 0x10) /* fixed volume */
		output = ch->regs[0] & 0x0F;
	else
		output = 15 - ch->envadr;

	if (ch->spd < (4 << CPS_BITS)) return 0;

	output = LinearToLog(output) + ch->mastervolume;
	output += (ch->rng & (1 << ((ch->regs[2] & 0x80) ? 3 : 12))) != 0;
	return LogToLinear(output, LOG_LIN_BITS - LIN_BITS - 16 + 1);
}

static void APUSoundDpcmStart(APU_DPCM *ch)
{
	ch->adr = 0xC000 + ((Uint)ch->regs[2] << 6);
	ch->length = (((Uint)ch->regs[3] << 4) + 1) << 3;
	ch->irq_report = 0; 
	ch->input = NES6502Read(ch->adr++);
	ch->first = 1;
	ch->output = 0;
}

static Int32 APUSoundDpcmRender(APU_DPCM *ch)
{
	if (ch->update)
	{
		if (ch->update & 1)
		{
			if (!(ch->regs[0] & 0x80)) ch->irq_report = 0;
			ch->spd = dpcm_freq[ch->regs[0] & 0x0F];
		}
		if (ch->update & 2)
		{
			if (ch->first)
			{
				if (ch->regs[1]) ch->first = 0;
			}
			else
			{
				ch->output = ((Int)(ch->regs[1] >> 1) & 0x3F) - 0x1F;
			}
		}
		ch->update = 0;
	}

	if (ch->key) {
		if (!ch->length) return 0;

		ch->cycles -= ch->cps;
		while (ch->cycles < 0)
		{
			ch->cycles += ch->spd;
			if (ch->length == 0) continue;
			if (ch->input & 1)
				ch->output += (ch->output < 0x1F);
			else
				ch->output -= (ch->output > -0x20);
			ch->input >>= 1;

			if (--ch->length == 0)
			{
				if (ch->regs[0] & 0x40)
				{
					APUSoundDpcmStart(ch);	/*loop */
				}
				else
				{
					if (ch->regs[0] & 0x80)
					{
						NES6502Irq();	/* irq gen */
						ch->irq_report = 0x80;
					}
					ch->first = 1;
					ch->output = 0;
					ch->length = 0;
				}
			}
			else if ((ch->length & 7) == 0)
			{
				ch->input = NES6502Read(ch->adr++);
			}
		}
	}

	if (ch->mute) return 0;

	return LogToLinear(LinearToLog(ch->output) + ch->mastervolume, LOG_LIN_BITS - LIN_BITS - 16);
}

static Int32 __fastcall APUSoundRender(void)
{
	Int32 accum = 0;
	accum += APUSoundSquareRender(&apu.square[0]);
	accum += APUSoundSquareRender(&apu.square[1]);
	accum += APUSoundTriangleRender(&apu.triangle);
	accum += APUSoundNoiseRender(&apu.noise);
	accum += APUSoundDpcmRender(&apu.dpcm);
	return accum;
}

static NES_AUDIO_HANDLER s_apu_audio_handler[] = {
	{ 1, APUSoundRender, }, 
	{ 0, 0, }, 
};

static void __fastcall APUSoundVolume(Uint volume)
{
	volume = (volume << (LOG_BITS - 8)) << 1;
	apu.square[0].mastervolume = volume;
	apu.square[1].mastervolume = volume;
	apu.triangle.mastervolume = volume;
	apu.noise.mastervolume = volume;
	apu.dpcm.mastervolume = volume;
}

static NES_VOLUME_HANDLER s_apu_volume_handler[] = {
	{ APUSoundVolume, },
	{ 0, }, 
};

static void __fastcall APUSoundWrite(Uint address, Uint value)
{
	if (0x4000 <= address && address <= 0x4015)
	{
		apu.regs[address - 0x4000] = value;
		switch (address)
		{
			case 0x4000: case 0x4001: case 0x4002: case 0x4003:
			case 0x4004: case 0x4005: case 0x4006: case 0x4007:
				{
					int ch = address >= 0x4004;
					int port = address & 3;
					apu.square[ch].regs[port] = value;
					apu.square[ch].update |= 1 << port;
				}
				break;
			case 0x4008:
				apu.triangle.regs[0] = value;
				apu.triangle.update |= 1 << 0; 
				if (apu.triangle.key & KEY_ON && !apu.triangle.counter_started)
				{
					apu.triangle.key &= ~KEY_RELEASE;
					apu.triangle.length2 = ((apu.triangle.regs[0] & 0x7F) * apu.triangle.freq) >> (6 + 2);
				}
				break;
			case 0x4009: case 0x400A:
				{
					int port = address & 3;
					apu.triangle.regs[port] = value;
					apu.triangle.update |= 1 << port;
				}
				break;
			case 0x400B:
				apu.triangle.regs[3] = value;
				apu.triangle.update |= 1 << 3; 
				apu.triangle.write_latency = 3;
				if (apu.triangle.key & KEY_ON)
				{
					apu.triangle.key &= ~KEY_RELEASE;
					apu.triangle.counter_started = 0;
					apu.triangle.length  = (vbl_length[apu.triangle.regs[3] >> 3] * apu.triangle.freq) >> 6;
					apu.triangle.length2 = ((apu.triangle.regs[0] & 0x7F) * apu.triangle.freq) >> (6 + 2);
				}
				break;
			case 0x400C: case 0x400D: case 0x400E: case 0x400F:
				{
					int port = address & 3;
					apu.noise.regs[port] = value;
					apu.noise.update |= 1 << port; 
				}
				break;
			case 0x4010: case 0x4011: case 0x4012: case 0x4013:
				{
					int port = address & 3;
					apu.dpcm.regs[port] = value;
					apu.dpcm.update |= 1 << port; 
				}
				break;
			case 0x4015:
				if (value & 1)
					apu.square[0].key |= KEY_ON;
				else
				{
					apu.square[0].key &= ~KEY_ON;
					apu.square[0].key |= KEY_RELEASE;
					apu.square[0].length = 0;
				}
				if (value & 2)
					apu.square[1].key |= KEY_ON;
				else
				{
					apu.square[1].key &= ~KEY_ON;
					apu.square[1].key |= KEY_RELEASE;
					apu.square[1].length = 0;
				}
				if (value & 4)
				{
					apu.triangle.key |= KEY_ON;
				}
				else
				{
					apu.triangle.key &= ~KEY_ON;
					apu.triangle.key |= KEY_RELEASE;
					apu.triangle.length = 0;
					apu.triangle.length2 = 0;
					apu.triangle.counter_started = 0;
					apu.triangle.write_latency = 0;
				}
				if (value & 8)
				{
					apu.noise.key = KEY_ON;
				}
				else
				{
					apu.noise.key = 0;
					apu.noise.length = 0;
				}
				if (value & 16)
				{
					if (!apu.dpcm.key)
					{
						apu.dpcm.key = KEY_ON;
						APUSoundDpcmStart(&apu.dpcm);
					}
				}
				else
				{
					if (apu.dpcm.key)
					{
						apu.dpcm.key = 0;
						apu.dpcm.output = 0;
					}
				}
				break;
		}
	}
}

static NES_WRITE_HANDLER s_apu_write_handler[] =
{
	{ 0x4000, 0x4017, APUSoundWrite, },
	{ 0,      0,      0, },
};

static Uint __fastcall APUSoundRead(Uint address)
{
	if (0x4000 <= address && address <= 0x4014)
		return apu.regs[address - 0x4000];
	if (0x4015 == address)
	{
#if 0
		int key = 0;
		if (apu.square[0].key == KEY_ON) key |= 1;
		if (apu.square[1].key == KEY_ON) key |= 2;
		if (apu.triangle.key == KEY_ON) key |= 4;
		if (apu.noise.key == KEY_ON) key |= 8;
		if (apu.dpcm.key == KEY_ON) key |= 16;
		return key | 0x40 | apu.dpcm.irq_report;
#else
		return (apu.regs[0x15] & 0x1F) | 0x40 | apu.dpcm.irq_report;
#endif
	}
	return 0xFF;
}

static NES_READ_HANDLER s_apu_read_handler[] =
{
	{ 0x4000, 0x4017, APUSoundRead, },
	{ 0,      0,      0, },
};

#include "sct_init.h"

static Uint32 DivFix(Uint32 p1, Uint32 p2, Uint32 fix)
{
	Uint32 ret;
	ret = p1 / p2;
	p1  = p1 % p2;/* p1 = p1 - p2 * ret; */
	while (fix--)
	{
		p1 += p1;
		ret += ret;
		if (p1 >= p2)
		{
			p1 -= p2;
			ret++;
		}
	}
	return ret;
}


static void APUSoundSquareReset(APU_SQUARE *ch)
{
	XMEMSET(ch, 0, sizeof(APU_SQUARE));
	ch->freq = NESAudioFrequencyGet();
	ch->cps = DivFix(NES_BASECYCLES, 12 * ch->freq, CPS_BITS);
}

static void APUSoundTriangleReset(APU_TRIANGLE *ch)
{
	XMEMSET(ch, 0, sizeof(APU_TRIANGLE));
	ch->freq = NESAudioFrequencyGet();
	ch->cps = DivFix(NES_BASECYCLES, 12 * ch->freq, CPS_BITS);
}

static void APUSoundNoiseReset(APU_NOISE *ch)
{
	XMEMSET(ch, 0, sizeof(APU_NOISE));
	ch->freq = NESAudioFrequencyGet();
	ch->cps = DivFix(NES_BASECYCLES, 12 * ch->freq, CPS_BITS);
}

static void APUSoundDpcmReset(APU_DPCM *ch)
{
	XMEMSET(ch, 0, sizeof(APU_DPCM));
	ch->cps = DivFix(NES_BASECYCLES, 12 * NESAudioFrequencyGet(), CPS_BITS);
}

static void __fastcall APUSoundReset(void)
{
	Uint i;
	static Uint8 initdata[0x14] = {
		0x10,0x00,0x00,0x00,0x10,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,0x10,0x00,0x00,0x00,
		0x00,0x00,0x00,0x00,
	};
	APUSoundSquareReset(&apu.square[0]);
	APUSoundSquareReset(&apu.square[1]);
	APUSoundTriangleReset(&apu.triangle);
	APUSoundNoiseReset(&apu.noise);
	APUSoundDpcmReset(&apu.dpcm);

	for (i = 0; i < 0x14; i++) APUSoundWrite(0x4000 + i, initdata[i]);
	APUSoundWrite(0x4015, 0x0F);

	apu.dpcm.first = 1;
	apu.dpcm.output = 0;
}

static NES_RESET_HANDLER s_apu_reset_handler[] = {
	{ NES_RESET_SYS_NOMAL, APUSoundReset, }, 
	{ 0,                   0, }, 
};

void APUSoundInstall(void)
{
	LogTableInitialize();
	NESAudioHandlerInstall(s_apu_audio_handler);
	NESVolumeHandlerInstall(s_apu_volume_handler);
	NESReadHandlerInstall(s_apu_read_handler);
	NESWriteHandlerInstall(s_apu_write_handler);
	NESResetHandlerInstall(s_apu_reset_handler);
}


/* ----------------- */
/*  MMC5 EXTEND RAM  */
/* ----------------- */

#include "sct_main.h"

static Uint8 mmc5exram[0x400];

static Uint __fastcall mmc5exram_read(Uint address)
{
	return mmc5exram[address & 0x03FF];
}

static void __fastcall mmc5exram_write(Uint address, Uint value)
{
	mmc5exram[address & 0x03FF] = value;
}

static NES_READ_HANDLER mmc5exram_read_handler[] =
{
	{ 0x5C00, 0x5FEF, mmc5exram_read, },
	{ 0,      0,      0, },
};
static NES_WRITE_HANDLER mmc5exram_write_handler[] =
{
	{ 0x5C00, 0x5FEF, mmc5exram_write, },
	{ 0,      0,      0, },
};

#include "sct_init.h"

void MMC5ExtendRamInstall(void)
{
	NESReadHandlerInstall(mmc5exram_read_handler);
	NESWriteHandlerInstall(mmc5exram_write_handler);
}



/* ----------------- */
/*  MMC5 MULTIPLIER  */
/* ----------------- */

#include "sct_main.h"

static Uint8 mmc5multiplier[2];

static Uint __fastcall mmc5mul_read(Uint address)
{
	Uint mul;
	address = address - 0x5205;
	/* if (address > 1) return; */
	mul = mmc5multiplier[0] * mmc5multiplier[1];
	return address ? (Uint8)(mul >> 8) : (Uint8)mul;
}

static void __fastcall mmc5mul_write(Uint address, Uint value)
{
	address = address - 0x5205;
	/* if (address > 1) return; */
	mmc5multiplier[address] = value;
}

static NES_READ_HANDLER mmc5mul_read_handler[] =
{
	{ 0x5205, 0x5206, mmc5mul_read, },
	{ 0,      0,      0, },
};
static NES_WRITE_HANDLER mmc5mul_write_handler[] =
{
	{ 0x5205, 0x5206, mmc5mul_write, },
	{ 0,      0,      0, },
};

#include "sct_init.h"

void MMC5MutiplierInstall(void)
{
	NESReadHandlerInstall(mmc5mul_read_handler);
	NESWriteHandlerInstall(mmc5mul_write_handler);
}



/* ------------ */
/*  MMC5 SOUND  */
/* ------------ */

#include "sct_main.h"

static MMC5SOUND mmc5;

static Int32 __fastcall MMC5SoundRender(void)
{
	Int32 accum = 0;
	accum += APUSoundSquareRender(&mmc5.square[0]);
	accum += APUSoundSquareRender(&mmc5.square[1]);
#if 0
	if (mmc5.da.key && !mmc5.da.mute) accum += mmc5.da.output << 8;
#endif
	return accum;
}

static NES_AUDIO_HANDLER s_mmc5_audio_handler[] = {
	{ 1, MMC5SoundRender, }, 
	{ 0, 0, }, 
};

static void __fastcall MMC5SoundVolume(Uint volume)
{
	volume = (volume << (LOG_BITS - 8)) << 1;
	mmc5.square[0].mastervolume = volume;
	mmc5.square[1].mastervolume = volume;
}

static NES_VOLUME_HANDLER s_mmc5_volume_handler[] = {
	{ MMC5SoundVolume, },
	{ 0, }, 
};

static void __fastcall MMC5SoundWrite(Uint address, Uint value)
{
	if (0x5000 <= address && address <= 0x5015)
	{
		switch (address)
		{
			case 0x5000: case 0x5002: case 0x5003:
			case 0x5004: case 0x5006: case 0x5007:
				{
					int ch = address >= 0x5004;
					int port = address & 3;
					mmc5.square[ch].regs[port] = value;
					mmc5.square[ch].update |= 1 << port; 
				}
				break;
			case 0x5011:
				mmc5.da.output = ((Int)value) - 0x80;
				break;
			case 0x5010:
				mmc5.da.key = (value & 0x01);
				break;
			case 0x5015:
				if (value & 1)
					mmc5.square[0].key = 1;
				else
				{
					mmc5.square[0].key = 0;
					mmc5.square[0].length = 0;
				}
				if (value & 2)
					mmc5.square[1].key = 1;
				else
				{
					mmc5.square[1].key = 0;
					mmc5.square[1].length = 0;
				}
				break;
		}
	}
}

static NES_WRITE_HANDLER s_mmc5_write_handler[] =
{
	{ 0x5000, 0x5015, MMC5SoundWrite, },
	{ 0,      0,      0, },
};

#include "sct_init.h"

static void __fastcall MMC5SoundDaReset(MMC5_DA *ch)
{
	XMEMSET(ch, 0, sizeof(MMC5_DA));
}

static void __fastcall MMC5SoundReset(void)
{
	APUSoundSquareReset(&mmc5.square[0]);
	APUSoundSquareReset(&mmc5.square[1]);
	MMC5SoundDaReset(&mmc5.da);

	MMC5SoundWrite(0x5000, 0x00);
	MMC5SoundWrite(0x5002, 0x00);
	MMC5SoundWrite(0x5003, 0x00);
	MMC5SoundWrite(0x5004, 0x00);
	MMC5SoundWrite(0x5006, 0x00);
	MMC5SoundWrite(0x5007, 0x00);
	MMC5SoundWrite(0x5010, 0x00);
	MMC5SoundWrite(0x5011, 0x00);
}

static NES_RESET_HANDLER s_mmc5_reset_handler[] = {
	{ NES_RESET_SYS_NOMAL, MMC5SoundReset, }, 
	{ 0,                   0, }, 
};

void MMC5SoundInstall(void)
{
	LogTableInitialize();
	NESAudioHandlerInstall(s_mmc5_audio_handler);
	NESVolumeHandlerInstall(s_mmc5_volume_handler);
	NESWriteHandlerInstall(s_mmc5_write_handler);
	NESResetHandlerInstall(s_mmc5_reset_handler);
}
