#include "../nestypes.h"
#include "../audiosys.h"
#include "../handler.h"
#include "logtable.h"
#include "s_vrc7.h"

#include <math.h>

#ifndef M_PI
#define M_PI			3.14159265358979323846
#endif

#define NES_BASECYCLES (21477270)
#define CPS_SHIFT 18

#define OPLL_CH_MAX 6/* 9 */

/* PG_BITS + PG_SHIFT < 32 */
#define PG_BITS 10
#define AM_BITS 8
#define PM_BITS 8

#define AM_DEPTH 4.8  /* 1.0 */		/* dB */
#define PM_DEPTH 14.0 /* 7.0 */		/* cent */
#define PM_SHIFT 16

#define PG_SHIFT (20 - PG_BITS)
#define EG_SHIFT 7/*8*/
#define LOG_KEYOFF (32 << LOG_BITS)
#define EG_KEYOFF (LOG_KEYOFF << EG_SHIFT)

typedef struct {
	Uint32	pg_phase;
	Uint32	pg_spd;
	Int32	vib_cycles;
	Uint32	input;
	Uint32	eg_phase;
	Uint32	eg_sl;
	Uint32	eg_arr;
	Uint32	eg_drr;
	Uint32	eg_rrr;
	Uint8	pg_vib;
	Uint8	tl;
	Uint8	dist;
	Uint8	eg_mode;
	Uint8	eg_type;
	Uint8	su_type;
	Uint8	eg_ar;
	Uint8	eg_dr;
	Uint8	eg_rr;
	Uint8	eg_ks;
	Uint8	eg_am;
} OPLL_OP;
typedef struct {
	Uint32	cps;
	Uint32	spd;
	Int32	cycles;
	Uint32	adr;
	Uint32	adrmask;
	Uint32	*table;
	Uint32	output;
} OPLL_LFO;

typedef enum {
	EGM_OFF,
	EGM_A,
	EGM_D,
	EGM_S,
	EGM_R
} OPLL_EGMODE;

typedef struct {
	Uint32 cps;
	Int32 cycles;
	Uint32 fbbuf[2];
	Uint32	output;
	OPLL_OP op[2];
	Uint mastervolume;
	Uint8 tone[8];
	Uint8 key;
	Uint8 toneno;
	Uint8 freql;
	Uint8 freqh;
	Uint8 fb;
	Uint8 update;
} OPLL_CH;

#define OPLL_UPDATE_FREQH		(1 << 0)
#define OPLL_UPDATE_FREQL		(1 << 1)
#define OPLL_UPDATE_TONE		(1 << 2)

typedef struct {
	OPLL_CH ch[OPLL_CH_MAX];
	OPLL_LFO lfo[2];
	Uint32 mastervolume;
	Uint8 usertone[8];
	Uint8 adr;
	Uint8 rhythmc;
	Uint8 toneupdate;
} OPLLSOUND;

#include "sct_main.h"

static OPLLSOUND ym2413s;
static Uint32 amtbl[1 << AM_BITS];
static Uint32 pmtbl[1 << PM_BITS];
static Uint32 sintbl[1 << PG_BITS];
static Uint32 sintbld[1 << PG_BITS];

static Uint8 multbl[16] =
{
	1,1*2,2*2,3*2,4*2,5*2,6*2,7*2,8*2,9*2,10*2,10*2,12*2,12*2,15*2,15*2,
};
static Uint8 fbtbl[8] =
{
	31,6,5,4,3,2,1,0,
};
static Uint8 oplltone[][8] =
{
#if 1
	/* VRCVII.TXT(Kevin Horton) someone's tone */
	/*  AM1  AM2  KS1  KS2  AD1  AD2  SR1  SR2 */
	{  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00 },
	{  0x31,0x22,0x23,0x07,0xF0,0xF0,0xE8,0xF7 },
	{  0x03,0x31,0x68,0x05,0xF2,0x74,0x79,0x9C },
	{  0x01,0x51,0x72,0x04,0xF1,0xD3,0x9D,0x8B },
	{  0x22,0x61,0x1B,0x05,0xC0,0xA1,0xF8,0xE8 },
	{  0x22,0x61,0x2C,0x03,0xD2,0xA1,0xA7,0xE8 },
	{  0x31,0x22,0xFA,0x01,0xF1,0xF1,0xF4,0xEE },
	{  0x21,0x61,0x28,0x06,0xF1,0xF1,0xCE,0x9B },
	{  0x27,0x61,0x60,0x00,0xF0,0xF0,0xFF,0xFD },
	{  0x60,0x21,0x2B,0x06,0x85,0xF1,0x79,0x9D },
	{  0x31,0xA1,0xFF,0x0A,0x53,0x62,0x5E,0xAF },
	{  0x03,0xA1,0x70,0x0F,0xD4,0xA3,0x94,0xBE },
	{  0x2B,0x61,0xE4,0x07,0xF6,0x93,0xBD,0xAC },
	{  0x21,0x63,0xED,0x07,0x77,0xF1,0xC7,0xE8 },
	{  0x21,0x61,0x2A,0x03,0xF3,0xE2,0xB6,0xD9 },
	{  0x21,0x63,0x37,0x03,0xF3,0xE2,0xB6,0xD9 },
#elif 0
	/*  AM1  AM2  KS1  KS2  AD1  AD2  SR1  SR2 MSX-MUSIC */
	{  0x00,0x00,0xF0,0x20,0x00,0x00,0xFF,0xFF }, /* @63 */
	{  0x61,0x61,0x12,0x20,0xB4,0x56,0x14,0x17 }, /* @02 */
	{  0x02,0x41,0x15,0x20,0xA3,0xA3,0x75,0x05 }, /* @10 */
	{  0x31,0x11,0x0E,0x20,0xD9,0xB2,0x11,0xF4 }, /* @00 */
	{  0x61,0x31,0x20,0x20,0x6C,0x43,0x18,0x26 }, /* @03 */
	{  0xA2,0x30,0xA0,0x20,0x88,0x54,0x14,0x06 }, /* @04 */
	{  0x31,0x34,0x20,0x20,0x72,0x56,0x0A,0x1C }, /* @05 */
	{  0x31,0x71,0x16,0x20,0x51,0x52,0x26,0x24 }, /* @06 */
	{  0xE1,0x63,0x0A,0x20,0xFC,0xF8,0x28,0x29 }, /* @09 */
	{  0x61,0x71,0x0D,0x20,0x75,0xF2,0x18,0x03 }, /* @48 */
	{  0x42,0x44,0x0B,0x20,0x94,0xB0,0x33,0xF6 }, /* @24 */
	{  0x01,0x00,0x06,0x20,0xA3,0xE2,0xF4,0xF4 }, /* @14 */
	{  0xF9,0xF1,0x24,0x20,0x95,0xD1,0xE5,0xF2 }, /* @16 */
	{  0x40,0x31,0x89,0x20,0xC7,0xF9,0x14,0x04 }, /* @23 */
	{  0x11,0x11,0x11,0x20,0xC0,0xB2,0x01,0xF4 }, /* @33 */
	{  0x23,0x43,0x09,0x20,0xDD,0xBF,0x4A,0x05 }, /* @12 */
#elif 0	/* NEZ temp */
	/*  AM1  AM2  KS1  KS2  AD1  AD2  SR1  SR2 */
	{  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00 },	/* @63 */
	{  0x70,0x20,0x1C,0x07,0x51,0xB4,0x03,0x67 },	/* @2 */
	{  0x13,0x11,0x93,0x85,0xFF,0xFF,0x21,0x03 },	/* @10 */
	{  0x21,0x31,0x4F,0x03,0xF2,0xD2,0x52,0x73 },	/* @00 */
	{  0x61,0x31,0x20,0x00,0x6C,0x43,0x18,0x26 },	/* @03! */
	{  0x32,0x61,0x9A,0x06,0x51,0xB2,0x1B,0x3B },	/* @04 */
	{  0x21,0x24,0x96,0x05,0xF0,0xC0,0x09,0x0A },	/* @05 */
	{  0x31,0x62,0x1C,0x87,0x41,0xC2,0x0B,0x3B },	/* @06 */
	{  0x01,0x31,0x4F,0x03,0xF0,0x90,0xFF,0xFF },	/* @09 */
	{  0xB1,0xA2,0xC5,0x01,0x6E,0xDB,0x17,0x0E },	/* @48 */
	{  0x31,0x62,0x1C,0x07,0x75,0x54,0x03,0x44 },	/* @24 */
	{  0x31,0x32,0x8E,0x85,0xF1,0xF3,0xF9,0xF9 },	/* @14 */
	{  0x93,0x91,0x97,0x87,0xAA,0xAC,0x12,0x21 },	/* @16 */
	{  0x11,0x31,0x05,0x05,0xF9,0xF1,0x25,0x34 },	/* @23 */
	{  0x11,0x11,0x11,0x00,0xC0,0xB2,0x01,0xF4 },	/* @33 */
	{  0x03,0x11,0x5E,0x07,0xF5,0xF2,0x71,0x83 },	/* @12 */
#endif

	{  0x13,0x11,0x25,0x00,0xd7,0xb7,0xf4,0xf4 }, /* 16 Rhythm 1: */
	{  0x13,0x11,0x25,0x00,0xd7,0xb7,0xf4,0xf4 }, /* 17 Rhythm 2: */
	{  0x13,0x11,0x25,0x00,0xd7,0xb7,0xf4,0xf4 }, /* 18 Rhythm 3: */
};

__inline static Uint32 PGCALC(OPLL_OP *pop)
{
	if (pop->pg_vib)
	{
		pop->vib_cycles -= 1 << PM_SHIFT;
		while (pop->vib_cycles < 0)
		{
			pop->vib_cycles += ym2413s.lfo[1].output;
			pop->pg_phase += pop->pg_spd;
		}
	}
	else
	{
		pop->pg_phase += pop->pg_spd;
	}
	return (pop->input + (pop->pg_phase >> PG_SHIFT)) & ((1 << PG_BITS) - 1);
}

__inline static void OPKEY(OPLL_OP *pop, Uint key)
{
	if (key)
	{
		pop->eg_mode = EGM_A;
		pop->eg_phase = 63 << (EG_SHIFT + LOG_BITS - 3);
	}
	else
	{
		if (pop->eg_mode == EGM_OFF) return;
		pop->eg_mode = EGM_R;
	}
}
__inline static Uint32 EGCALC(OPLL_OP *pop)
{
	Uint32 egout = pop->tl << (LOG_BITS - 3);
	if (pop->eg_am) egout += ym2413s.lfo[0].output;
	switch (pop->eg_mode)
	{
		case EGM_A:
			{
				Uint32 attack;
				attack = 1 + (pop->eg_phase >> (15 - (pop->eg_arr >> 2)));
				if (pop->eg_phase > attack)
				{
					pop->eg_phase -= attack;
					break;
				}
			}
			pop->eg_phase = 0;
			pop->eg_mode = EGM_D;
		case EGM_D:
			if (pop->eg_phase < (pop->eg_sl << (EG_SHIFT + LOG_BITS - 3)))
			{
				pop->eg_phase += pop->eg_drr;
				break;
			}
			pop->eg_mode = EGM_S;
		case EGM_S:
			if (pop->eg_type) break;
			if (pop->eg_phase < EG_KEYOFF)
			{
				pop->eg_phase += pop->eg_rrr;
			}
			break;
		case EGM_R:
			if (pop->eg_phase < EG_KEYOFF)
			{
				if (pop->su_type)
				{
					pop->eg_phase += (5 << 2);
				}
				else if (pop->eg_type)
					pop->eg_phase += pop->eg_rrr;
				else
					pop->eg_phase += pop->eg_rrr << 1;
				break;
			}
			pop->eg_mode = EGM_OFF;
		case EGM_OFF:
			return EG_KEYOFF;
	}
	return ((pop->eg_phase >> EG_SHIFT) + egout) << 1;
}

static Uint32 OPCALC(OPLL_OP *pop)
{
	Uint32 sg;
	Uint32 *sintblp = pop->dist ? sintbld : sintbl;
	sg = sintblp[PGCALC(pop)];
	return EGCALC(pop) + sg;
}

__inline static void EGSET(OPLL_CH *pch)
{
	pch->op[0].eg_am = pch->tone[0] & 0x80;
	pch->op[1].eg_am = pch->tone[1] & 0x80;
	pch->op[0].tl = pch->tone[2] & 0x3F;
	pch->op[0].dist = pch->tone[3] & 0x08;
	pch->op[1].dist = pch->tone[3] & 0x10;
	pch->op[0].eg_ks = pch->tone[2] >> 6;
	pch->op[1].eg_ks = pch->tone[3] >> 6;
	pch->op[0].eg_type = pch->tone[0] & 0x20;
	pch->op[1].eg_type = pch->tone[1] & 0x20;
	pch->op[0].eg_ar = (pch->tone[4] & 0xF0) >> 4;
	pch->op[1].eg_ar = (pch->tone[5] & 0xF0) >> 4;
	pch->op[0].eg_dr = (pch->tone[4] & 0x0F);
	pch->op[1].eg_dr = (pch->tone[5] & 0x0F);
	pch->op[0].eg_rr = (pch->tone[6] & 0x0F);
	pch->op[1].eg_rr = (pch->tone[7] & 0x0F);
	pch->op[0].eg_sl = (pch->tone[6] & 0xF0) >> (4 - 2);
	pch->op[1].eg_sl = (pch->tone[7] & 0xF0) >> (4 - 2);
}
__inline static void EGSET2(OPLL_OP *pop)
{
	pop->eg_arr = pop->eg_ar << 2;
	pop->eg_drr = pop->eg_dr << 2;
	pop->eg_rrr = pop->eg_rr << 2;
}

__inline static void EGSET2KSR(OPLL_OP *pop, Uint freqh)
{
	int ksr = (freqh & 0xF) >> (3 - pop->eg_ks);
	if (pop->eg_ar)
	{
		pop->eg_arr = (pop->eg_ar << 2) + ksr;
		if (pop->eg_arr > 63) pop->eg_arr = 63;
	}
	else
	{
		pop->eg_arr = pop->eg_ar << 2;
	}
	if (pop->eg_dr)
	{
		pop->eg_drr = (pop->eg_dr << 2) + ksr;
		if (pop->eg_drr > 63) pop->eg_drr = 63;
	}
	else
	{
		pop->eg_drr = pop->eg_dr << 2;
	}
	if (pop->eg_rr)
	{
		pop->eg_rrr = (pop->eg_rr << 2) + ksr;
		if (pop->eg_rrr > 63) pop->eg_rrr = 63;
	}
	else
	{
		pop->eg_rrr = pop->eg_rr << 2;
	}
}

__inline static void PGSET(OPLL_CH *pch)
{
	int fnum, block;
	fnum = ((pch->freqh << 8) + pch->freql) & 0x1FF;
	block = (pch->freqh >> 1) & 7;
	pch->op[0].pg_spd = (multbl[pch->tone[0] & 0xF] * fnum) << block;
	pch->op[1].pg_spd = (multbl[pch->tone[1] & 0xF] * fnum) << block;
}

__inline static Int32 CHCALC(OPLL_CH *pch)
{
	if (pch->update)
	{
		if (pch->update & (OPLL_UPDATE_FREQL | OPLL_UPDATE_FREQH | OPLL_UPDATE_TONE))
		{
			PGSET(pch);
		}
		if (pch->update & OPLL_UPDATE_TONE)
		{
			pch->fb = fbtbl[pch->tone[3] & 0x7] + (LOG_LIN_BITS - PG_BITS);
			pch->fbbuf[0] = pch->fbbuf[1] = 0;
			if (pch->fb > 31) pch->fb = 31;
			pch->op[0].pg_vib = pch->tone[0] & 0x40;
			pch->op[1].pg_vib = pch->tone[1] & 0x40;
			EGSET(pch);
		}
		if (pch->update & OPLL_UPDATE_FREQH)
		{
			if ((pch->key ^ pch->freqh) & 0x10)
			{
				pch->key ^= 0x10;
				OPKEY(&pch->op[0], (pch->key & 0x10));
				OPKEY(&pch->op[1], (pch->key & 0x10));
			}
			pch->op[0].su_type = pch->freqh & 0x20;
			pch->op[1].su_type = pch->freqh & 0x20;
		}
		if (pch->update & (OPLL_UPDATE_FREQH | OPLL_UPDATE_TONE))
		{
			int op;
			for (op = 0; op < 2; op++)
			{
				if (pch->tone[op] & 0x10)
					EGSET2KSR(&pch->op[op], pch->freqh);
				else
					EGSET2(&pch->op[op]);
			}
		}
		pch->update = 0;
	}
	pch->cycles -= pch->cps;
	while (pch->cycles < 0)
	{
		Uint opout;
		pch->cycles += 1 << CPS_SHIFT;
		pch->op[0].input = pch->fbbuf[0] + pch->fbbuf[1];
		opout = OPCALC(&pch->op[0]);
		pch->fbbuf[1] = pch->fbbuf[0];
		pch->fbbuf[0] = LogToLinear(opout, pch->fb + 1);
		pch->op[1].input = LogToLinear(opout, (LOG_LIN_BITS - PG_BITS - 1));
		pch->output = OPCALC(&pch->op[1]) + ym2413s.mastervolume;
	}
	return LogToLinear(pch->output, (LOG_LIN_BITS - 21));
}

__inline static Int32 LFOCALC(OPLL_LFO *pch)
{
	pch->cycles -= pch->cps;
	while (pch->cycles < 0)
	{
		pch->cycles += pch->spd;
		pch->adr++;
	}
	pch->adr &= pch->adrmask;
	pch->output = pch->table[pch->adr];
	return pch->output;
}

static Int32 __fastcall OPLLSoundRender(void)
{
	Int32 accum = 0;
	Uint ch;
#if OPLL_CH_MAX == 9
	Uint chmax = (ym2413s.rhythmc & 0x20) ? 6 : 9;
#else
	const Uint chmax = 6;	/* VRCVII */
#endif
	for (ch = 0; ch < 2; ch++) LFOCALC(&ym2413s.lfo[ch]);
	for (ch = 0; ch < chmax; ch++)
	{
		accum += CHCALC(&ym2413s.ch[ch]);
	}
	return accum;
}

static NES_AUDIO_HANDLER s_vrc7_audio_handler[] = {
	{ 1, OPLLSoundRender, }, 
	{ 0, 0, }, 
};

static void __fastcall OPLLSoundVolume(Uint volume)
{
	ym2413s.mastervolume = (volume << (LOG_BITS - 8)) << 1;
}

static NES_VOLUME_HANDLER s_vrc7_volume_handler[] = {
	{ OPLLSoundVolume, },
	{ 0, }, 
};

static void __fastcall OPLLSoundWriteAddr(Uint address, Uint value)
{
	ym2413s.adr = value;
}
static void __fastcall OPLLSoundWriteData(Uint address, Uint value)
{
	int ch = ym2413s.adr & 0xF;
	switch (ym2413s.adr & 0xF8)
	{
		case 0x00:
			ym2413s.usertone[ym2413s.adr] = value;
			ym2413s.toneupdate = ~0;
			break;
#if OPLL_CH_MAX == 9
		case 0x08:
			if (ch != 0xE) break;
			ym2413s.rhythmc = value;
			/* rhythmmode */
			break;
		case 0x18:
			if (ch >= OPLL_CH_MAX) break;
#endif
		case 0x10:
			ym2413s.ch[ch].freql = value;
			ym2413s.ch[ch].update |= OPLL_UPDATE_FREQL;
			break;
#if OPLL_CH_MAX == 9
		case 0x28:
			if (ch >= OPLL_CH_MAX) break;
#endif
		case 0x20:
			ym2413s.ch[ch].freqh = value;
			ym2413s.ch[ch].update |= OPLL_UPDATE_FREQH;
			break;
#if OPLL_CH_MAX == 9
		case 0x38:
			if (ch >= OPLL_CH_MAX) break;
#endif
		case 0x30:
#if OPLL_CH_MAX == 9
			if ((ym2413s.rhythmc & 0x20) && (ch > 5))
			{
				if (ch != 6)
					ym2413s.ch[ch].op[0].tl = ((value >> 4) & 0xF) << 2;
				ym2413s.ch[ch].op[1].tl = (value & 0xF) << 2;
			}
			else
#endif
			{
				if(ym2413s.ch[ch].toneno != (value >> 4))
				{
					if (value >> 4)
					{
						ym2413s.ch[ch].toneno = value >> 4;
						XMEMCPY(ym2413s.ch[ch].tone, oplltone[value >> 4], 8);
						ym2413s.ch[ch].update |= OPLL_UPDATE_TONE;
					}
					else if ((ym2413s.ch[ch].toneno != 0x10) || (ym2413s.toneupdate & (1 << ch)))
					{
						ym2413s.toneupdate &= ~(1 << ch);
						ym2413s.ch[ch].toneno = 0x10;
						XMEMCPY(ym2413s.ch[ch].tone, ym2413s.usertone, 8);
						ym2413s.ch[ch].update |= OPLL_UPDATE_TONE;
					}
				}
				ym2413s.ch[ch].op[1].tl = (value & 0xF) << 2;
			}
			break;
	}
}

static NES_WRITE_HANDLER s_vrc7_write_handler[] =
{
	{ 0x9010, 0x9010, OPLLSoundWriteAddr, },
	{ 0x9030, 0x9030, OPLLSoundWriteData, },
	{ 0,      0,      0, },
};

#include "sct_init.h"

static Uint32 DivFix(Uint32 p1, Uint32 p2, Uint fix)
{
	Uint32 ret;
	ret = p1 / p2;
	p1  = p1 % p2;/* p1 = p1 - p2 * ret; */
	while (fix--)
	{
		p1 += p1;
		ret += ret;
		if (p1 >= p2)
		{
			p1 -= p2;
			ret++;
		}
	}
	return ret;
}

static void __fastcall OPLLSoundReset(void)
{
	Uint i, w;
	double a;
	Uint32 cps;
	XMEMSET(&ym2413s, 0, sizeof(OPLLSOUND));
	cps = DivFix(NES_BASECYCLES, 6 * 72 * NESAudioFrequencyGet(), CPS_SHIFT);
	for (i = 0; i < OPLL_CH_MAX; i++)
	{
		ym2413s.ch[i].cps = cps;
		ym2413s.ch[i].toneno = 0x10;
		ym2413s.ch[i].fb = 31;
		ym2413s.ch[i].output = 0 + (LOG_KEYOFF << 1);
	}

	w = 1 << (PG_BITS - 2);
	sintbl[w * 0] = 0 + (LOG_KEYOFF << 1);
	sintbl[w * 1] = 0;
	sintbl[w * 2] = 1 + (LOG_KEYOFF << 1);
	sintbl[w * 3] = 1;
	sintbld[w * 0] = 0 + (LOG_KEYOFF << 1);
	sintbld[w * 1] = 0;
	sintbld[w * 2] = 0 + (LOG_KEYOFF << 1);
	sintbld[w * 3] = 0 + (LOG_KEYOFF << 1);
	for (i = 1; i < w; i++)
	{
		Uint32 ua;
		a = sin((2 * M_PI * i) / (double)(w << 2)) * (1 << LOG_LIN_BITS);
		ua = (Uint32)((LOG_LIN_BITS - (log(a) / log(2))) * (1 << LOG_BITS));
		sintbl[w * 0 + i] = sintbl[w * 2 - i] = 0 + (ua << 1);
		sintbl[w * 2 + i] = sintbl[w * 4 - i] = 1 + (ua << 1);
		sintbld[w * 0 + i] = sintbld[w * 2 - i] = 0 + (ua << 1);
		sintbld[w * 2 + i] = sintbld[w * 4 - i] = 0 + (LOG_KEYOFF << 1);
	}
	/* AM(tremoro) */
	w = 1 << AM_BITS;
	for (i = 0; i < w; i++)
	{
		a = (1 + sin((2 * M_PI * i) / w)) * ((1 << LOG_BITS) * AM_DEPTH / 12);
		amtbl[i] = (Uint32)a;
	}
	ym2413s.lfo[0].cps = cps;
	ym2413s.lfo[0].spd = DivFix(NES_BASECYCLES, (Uint32)((6 * 72 * (1 << AM_BITS)) * 3.7), CPS_SHIFT);
	ym2413s.lfo[0].adrmask = w - 1;
	ym2413s.lfo[0].table = amtbl;
	ym2413s.lfo[0].output = ym2413s.lfo[0].table[0];
	/* VIB */
	w = 1 << PM_BITS;
	for (i = 0; i < w; i++)
	{
		a = pow(2, PM_DEPTH * sin((2 * M_PI * i) / w) / 1200.0) * (1 << PM_SHIFT);
		pmtbl[i] = (Uint32)a;
	}
	ym2413s.lfo[1].cps = cps;
	ym2413s.lfo[1].spd = DivFix(NES_BASECYCLES, (Uint32)((6 * 72 * (1 << PM_BITS)) * 6.4), CPS_SHIFT);
	ym2413s.lfo[1].adrmask = w - 1;
	ym2413s.lfo[1].table = pmtbl;
	ym2413s.lfo[1].output = ym2413s.lfo[1].table[0];
}

static NES_RESET_HANDLER s_vrc7_reset_handler[] = {
	{ NES_RESET_SYS_NOMAL, OPLLSoundReset, }, 
	{ 0,                   0, }, 
};

void VRC7SetTone(Uint8 *p)
{
	int i, j;
	for (i = 1; i < 16; i++)
	{
		for (j = 0; j < 8; j++)
		{
			oplltone[i][j] = *p++;
		}
	}
}

void VRC7SoundInstall(void)
{
	LogTableInitialize();
	NESAudioHandlerInstall(s_vrc7_audio_handler);
	NESVolumeHandlerInstall(s_vrc7_volume_handler);
	NESWriteHandlerInstall(s_vrc7_write_handler);
	NESResetHandlerInstall(s_vrc7_reset_handler);
}
